@@ -36,7 +36,9 @@ use constant IS_MOD_PERL_2_BUILD => IS_MOD_PERL_2 &&
 use constant IS_APACHE_TEST_BUILD =>
     grep { -e "$_/lib/Apache/TestConfig.pm" } qw(Apache-Test . ..);
 
-use Symbol ();
+use constant CUSTOM_CONFIG_FILE => 'Apache/TestConfigData.pm';
+
+use lib ();
 use File::Copy ();
 use File::Find qw(finddepth);
 use File::Basename qw(dirname);
@@ -44,16 +46,29 @@ use File::Path ();
 use File::Spec::Functions qw(catfile abs2rel splitdir canonpath
                              catdir file_name_is_absolute devnull);
 use Cwd qw(fastcwd);
+use Socket ();
+use Symbol ();
 
 use Apache::TestConfigPerl ();
 use Apache::TestConfigParse ();
 use Apache::TestTrace;
 use Apache::TestServer ();
 use Apache::TestRun ();
-use Socket ();
 
 use vars qw(%Usage);
 
+# variables stored in $Apache::TestConfigData::vars
+my @data_vars_must = qw(httpd apxs);
+my @data_vars_opt  = qw(user group port);
+# mapping from $Apache::TestConfigData::vars to $ENV settings
+my %vars_to_env = (
+    httpd => 'APACHE_TEST_HTTPD',
+    apxs  => 'APACHE_TEST_APXS',
+    user  => 'APACHE_TEST_USER',
+    group => 'APACHE_TEST_GROUP',
+    port  => 'APACHE_TEST_PORT',
+);
+
 %Usage = (
    top_dir         => 'top-level directory (default is $PWD)',
    t_dir           => 'the t/ test directory (default is $top_dir/t)',
@@ -83,7 +98,7 @@ use vars qw(%Usage);
    sslcaorg        => 'SSL CA organization to use for tests (default is asf)',
    libmodperl      => 'path to mod_perl\'s .so (full or relative to LIBEXECDIR)',
    defines         => 'values to add as -D defines (for example, "VAR1 VAR2")',
-   (map { $_ . '_module_name', "$_ module name"} qw(cgi ssl thread access auth)),
+   (map { $_ . '_module_name', "$_ module name"} qw(cgi ssl thread access auth php)),
 );
 
 my %filepath_conf_opts = map { $_ => 1 }
@@ -162,6 +177,11 @@ sub modperl_2_inc_fixup {
 }
 
 sub modperl_build_config {
+
+    # we don't want to get mp2 preconfigured data in order to be able
+    # to get the interactive tests running.
+    return undef if $ENV{APACHE_TEST_INTERACTIVE_CONFIG_TEST};
+
     eval {
         require Apache::Build;
     } or return undef;
@@ -173,8 +193,11 @@ sub new_test_server {
     Apache::TestServer->new($args || $self)
 }
 
+# setup httpd-independent components
+# for httpd-specific call $self->httpd_config()
 sub new {
     my $class = shift;
+
     my $args;
 
     $args = shift if $_[0] and ref $_[0];
@@ -189,8 +212,18 @@ sub new {
         $args->{$key} = $val;
     }
 
-    my $thaw = {};
+    my $top_dir = fastcwd;
+    $top_dir = pop_dir($top_dir, 't');
+    # untaint as we are going to use it a lot later on in -T sensitive
+    # operations (.e.g @INC)
+    $top_dir = $1 if $top_dir =~ /(.*)/;
 
+    # make sure that t/conf/apache_test_config.pm is found
+    # (unfortunately sometimes we get thrown into / by Apache so we
+    # can't just rely on $top_dir
+    lib->import($top_dir);
+
+    my $thaw = {};
     #thaw current config
     for (qw(conf t/conf)) {
         last if eval {
@@ -201,7 +234,7 @@ sub new {
             #something else, which we can't be sure how to load
             bless $thaw, 'Apache::TestConfig';
         };
-    };
+    }
 
     if ($args->{thaw} and ref($thaw) ne 'HASH') {
         #dont generate any new config
@@ -220,7 +253,7 @@ sub new {
 
     # custom config options from Apache::TestConfigData
     # again, this should force reconfiguration
-    Apache::TestRun::custom_config_add_conf_opts($args);
+    custom_config_add_conf_opts($args);
 
     my $self = bless {
         clean => {},
@@ -245,8 +278,7 @@ sub new {
         $self->{$_} = delete $args->{$_};
     }
 
-    $vars->{top_dir} ||= fastcwd;
-    $vars->{top_dir} = pop_dir($vars->{top_dir}, 't');
+    $vars->{top_dir} ||= $top_dir;
 
     $self->add_inc;
 
@@ -282,15 +314,93 @@ sub new {
     $vars->{serveradmin}  ||= $self->default_serveradmin;
 
     $vars->{minclients}   ||= 1;
+    $vars->{maxclients_preset} = $vars->{maxclients} || 0;
+    # if maxclients wasn't explicitly passed try to
     # prevent 'server reached MaxClients setting' errors
     $vars->{maxclients}   ||= $vars->{minclients} + 1;
 
+    # if a preset maxclients valus is smaller than minclients,
+    # maxclients overrides minclients
+    if ($vars->{maxclients_preset} &&
+        $vars->{maxclients_preset} < $vars->{minclients}) {
+        $vars->{minclients} = $vars->{maxclients_preset};
+    }
+
+    # for threaded mpms MaxClients must be a multiple of
+    # ThreadsPerChild (i.e. maxclients % minclients == 0)
+    # so unless -maxclients was explicitly specified use a double of
+    # minclients
+    $vars->{maxclientsthreadedmpm} =
+        $vars->{maxclients_preset} || $vars->{minclients} * 2;
+
     $vars->{proxy}        ||= 'off';
     $vars->{proxyssl_url} ||= '';
     $vars->{defines}      ||= '';
 
+    $self->{hostport} = $self->hostport;
+    $self->{server} = $self->new_test_server;
+
+    return $self;
+
+}
+
+# figure out where httpd is and run extra config hooks which require
+# knowledge of where httpd is
+sub httpd_config {
+    my $self = shift;
+
     $self->configure_apxs;
     $self->configure_httpd;
+
+    my $vars = $self->{vars};
+    unless ($vars->{httpd} or $vars->{apxs}) {
+
+        # mod_perl 2.0 build (almost) always knows the right httpd
+
+        # location (and optionally apxs). if we get here we can't
+        # continue because the interactive config can't work with
+        # mod_perl 2.0 build (by design)
+        if (IS_MOD_PERL_2_BUILD){
+            my $mp2_build = modperl_build_config();
+            # if mod_perl 2 was built against the httpd source it
+            # doesn't know where to find apxs/httpd, so in this case
+            # fall back to interactive config
+            unless ($mp2_build->{MP_APXS}) {
+                die "mod_perl 2 was built against Apache sources, we " .
+                "don't know where httpd/apxs executables are, therefore " .
+                "skipping the test suite execution"
+            }
+
+            # not sure what else could go wrong but we can't continue
+            die "something is wrong, mod_perl 2.0 build should have " .
+                "supplied all the needed information to run the tests. " .
+                "Please post lib/Apache/BuildConfig.pm along with the " .
+                "bug report";
+        }
+
+        if ($ENV{APACHE_TEST_NO_STICKY_PREFERENCES}) {
+            error "You specified APACHE_TEST_NO_STICKY_PREFERENCES=1 " .
+                "in which case you must explicitly specify -httpd " .
+                "and/or -apxs options";
+            Apache::TestRun::exit_perl(0);
+        }
+
+        $self->clean(1);
+        # this method restarts the whole program via exec
+        # so it never returns
+        $self->custom_config_first_time($self->{vars});
+    }
+    else {
+        debug "Using httpd: $vars->{httpd}";
+    }
+
+    # if we have gotten that far we know at least about the location
+    # of httpd and or apxs, so let's save it if we haven't saved any
+    # custom configs yet
+    unless (custom_config_exists()) {
+        $self->custom_config_save($self->{vars});
+    }
+
     $self->inherit_config; #see TestConfigParse.pm
     $self->configure_httpd_eapi; #must come after inherit_config
 
@@ -299,10 +409,9 @@ sub new {
     $self->default_module(ssl    => [qw(mod_ssl)]);
     $self->default_module(access => [qw(mod_access mod_authz_host)]);
     $self->default_module(auth   => [qw(mod_auth mod_auth_basic)]);
+    $self->default_module(php    => [qw(sapi_apache2 mod_php4 mod_php5)]);
 
-    $self->{hostport} = $self->hostport;
-
-    $self->{server} = $self->new_test_server;
+    $self->{server}->post_config;
 
     $self;
 }
@@ -349,6 +458,8 @@ sub configure_httpd {
     my $self = shift;
     my $vars = $self->{vars};
 
+    debug "configuring httpd";
+
     $vars->{target} ||= (WIN32 ? 'Apache.EXE' : 'httpd');
 
     unless ($vars->{httpd}) {
@@ -408,8 +519,10 @@ sub configure_proxy {
 
     #if we proxy to ourselves, must bump the maxclients
     if ($vars->{proxy} =~ /^on$/i) {
-        $vars->{minclients}++;
-        $vars->{maxclients}++;
+        unless ($vars->{maxclients_preset}) {
+            $vars->{minclients}++;
+            $vars->{maxclients}++;
+        }
         $vars->{proxy} = $self->{vhosts}->{'mod_proxy'}->{hostport};
         return $vars->{proxy};
     }
@@ -699,7 +812,8 @@ sub hostport {
 sub find_apache_module {
     my($self, $module) = @_;
 
-    die "find_apache_module: module name argument is required" unless $module;
+    die "find_apache_module: module name argument is required"
+        unless $module;
 
     my $vars = $self->{vars};
     my $sroot = $vars->{serverroot};
@@ -874,6 +988,11 @@ sub perlscript_header {
         }
     }
 
+    {
+        my $dir = canonpath catdir $FindBin::Bin, "t", "lib";
+        push @dirs, $dir if -d $dir;
+    }
+
     my $dirs = join("\n    ", '', @dirs) . "\n";;
 
     return <<"EOF";
@@ -934,7 +1053,11 @@ sub makepath {
 sub open_cmd {
     my($self, $cmd) = @_;
     # untaint some %ENV fields
-    local @ENV{ qw(PATH IFS CDPATH ENV BASH_ENV) };
+    local @ENV{ qw(IFS CDPATH ENV BASH_ENV) };
+    local $ENV{PATH} = untaint_path($ENV{PATH});
+
+    # launder for -T
+    $cmd = $1 if $cmd =~ /(.*)/;
 
     my $handle = Symbol::gensym();
     open $handle, "$cmd|" or die "$cmd failed: $!";
@@ -1090,6 +1213,18 @@ sub parse_vhost {
     };
 }
 
+sub find_and_load_module {
+    my ($self, $name) = @_;
+    my $mod_path = $self->find_apache_module($name) or return;
+    my ($sym) = $name =~ m/mod_(\w+)\./;
+
+    if ($mod_path && -e $mod_path) {
+        $self->preamble(IfModule => "!mod_$sym.c",
+                        qq{LoadModule ${sym}_module "$mod_path"\n});
+    }
+    return 1;
+}
+
 sub replace_vhost_modules {
     my $self = shift;
 
@@ -1143,11 +1278,7 @@ sub generate_types_config {
 
     # handle the case when mod_mime is built as a shared object
     # but wasn't included in the system-wide httpd.conf
-    my $mod_mime = $self->find_apache_module('mod_mime.so');
-    if ($mod_mime && -e $mod_mime) {
-        $self->preamble(IfModule => '!mod_mime.c',
-                        qq{LoadModule mime_module "$mod_mime"\n});
-    }
+    $self->find_and_load_module('mod_mime.so');
 
     unless ($self->{inherit_config}->{TypesConfig}) {
         my $types = catfile $self->{vars}->{t_conf}, 'mime.types';
@@ -1193,8 +1324,10 @@ sub check_vars {
         }
 
         if ($vars->{proxyssl_url}) {
-            $vars->{minclients}++;
-            $vars->{maxclients}++;
+            unless ($vars->{maxclients_preset}) {
+                $vars->{minclients}++;
+                $vars->{maxclients}++;
+            }
         }
     }
 }
@@ -1239,25 +1372,21 @@ sub generate_extra_conf {
         push @extra_conf, $generated;
     }
 
-    # if at least one .in file was modified or the derivative is
-    # missing, regenerate them all (so information like assigned port
-    # numbers will be correct)
-    if ($self->extra_conf_files_needing_update) {
-        for my $file (@conf_files) {
-            local $Apache::TestConfig::File = $file;
+    # regenerate .conf files
+    for my $file (@conf_files) {
+        local $Apache::TestConfig::File = $file;
 
-            my $in = Symbol::gensym();
-            open($in, $file) or next;
+        my $in = Symbol::gensym();
+        open($in, $file) or next;
 
-            (my $generated = $file) =~ s/\.in$//;
-            my $out = $self->genfile($generated, $file);
-            $self->replace_vars($in, $out);
+        (my $generated = $file) =~ s/\.in$//;
+        my $out = $self->genfile($generated, $file);
+        $self->replace_vars($in, $out);
 
-            close $in;
-            close $out;
+        close $in;
+        close $out;
 
-            $self->check_vars;
-        }
+        $self->check_vars;
     }
 
     #we changed order to give ssl the first port after DEFAULT_PORT
@@ -1302,6 +1431,11 @@ sub sslca_generate {
 sub sslca_clean {
     my $self = shift;
 
+    # XXX: httpd config is required, for now just skip ssl clean if
+    # there is none. should probably add some flag which will tell us
+    # when httpd_config was already run
+    return unless $self->{vars}->{httpd} && $self->{vars}->{ssl_module};
+
     return unless $self->sslca_can;
 
     Apache::TestSSLCA::clean($self);
@@ -1392,6 +1526,8 @@ sub generate_httpd_conf {
 
     my $out = $self->genfile($conf_file);
 
+    $self->find_and_load_module('mod_alias.so');
+
     $self->preamble_run($out);
 
     for my $name (qw(user group)) { #win32/cygwin do not support
@@ -1413,14 +1549,6 @@ sub generate_httpd_conf {
 
     # handle the case when mod_alias is built as a shared object
     # but wasn't included in the system-wide httpd.conf
-    my $mod_alias = $self->find_apache_module('mod_alias.so');
-    if ($mod_alias && -e $mod_alias) {
-        print $out <<EOF;
-<IfModule !mod_alias.c>
-    LoadModule alias_module "$mod_alias"
-</IfModule>
-EOF
-    }
 
     print $out "<IfModule mod_alias.c>\n";
     for (keys %aliases) {
@@ -1516,7 +1644,13 @@ sub which {
 
     return undef unless $program;
 
-    for my $base (map { catfile($_, $program) } File::Spec->path()) {
+    my @dirs = File::Spec->path();
+
+    require Config;
+    my $perl_bin = $Config::Config{bin} || '';
+    push @dirs, $perl_bin if $perl_bin and -d $perl_bin;
+
+    for my $base (map { catfile $_, $program } @dirs) {
         if ($ENV{HOME} and not WIN32) {
             # only works on Unix, but that's normal:
             # on Win32 the shell doesn't have special treatment of '~'
@@ -1536,19 +1670,38 @@ sub which {
 sub apxs {
     my($self, $q, $ok_fail) = @_;
     return unless $self->{APXS};
-    my $devnull = devnull();
-    my $apxs = shell_ready($self->{APXS});
-    my $val = qx($apxs -q $q 2>$devnull);
-    chomp $val if defined $val; # apxs post-2.0.40 adds a new line
-    unless ($val) {
-        if ($ok_fail) {
-            return "";
+    my $val;
+    unless (exists $self->{_apxs}{$q}) {
+        local @ENV{ qw(IFS CDPATH ENV BASH_ENV) };
+        local $ENV{PATH} = untaint_path($ENV{PATH});
+        my $devnull = devnull();
+        my $apxs = shell_ready($self->{APXS});
+        $val = qx($apxs -q $q 2>$devnull);
+        chomp $val if defined $val; # apxs post-2.0.40 adds a new line
+        if ($val) {
+            $self->{_apxs}{$q} = $val;
         }
-        else {
-            warn "APXS ($self->{APXS}) query for $q failed\n";
+        unless ($val) {
+            if ($ok_fail) {
+                return "";
+            }
+            else {
+                warn "APXS ($self->{APXS}) query for $q failed\n";
+                return $val;
+            }
         }
     }
-    $val;
+    $self->{_apxs}{$q};
+}
+
+# Temporarily untaint PATH
+sub untaint_path {
+    my $path = shift;
+    ($path) = ( $path =~ /(.*)/ );
+    # win32 uses ';' for a path separator, assume others use ':'
+    my $sep = WIN32 ? ';' : ':';
+    # -T disallows relative and empty directories in the PATH
+    return join $sep, grep !/^(\.|$)/, split /$sep/, $path;
 }
 
 sub pop_dir {
@@ -1659,6 +1812,7 @@ sub as_string {
 
     # httpd opts
     my $test_config = Apache::TestConfig->new({thaw=>1});
+    # XXX: need to run httpd config to get the value of httpd
     if (my $httpd = $test_config->{vars}->{httpd}) {
         $httpd = shell_ready($httpd);
         $command = "$httpd -V";
@@ -1681,10 +1835,515 @@ sub as_string {
 # escape quotes)
 sub shell_ready {
     my $arg = shift;
-    $arg =~ s/"/\"/g;
+    $arg =~ s!\\?"!\\"!g;
     return qq["$arg"];
 }
 
+### Permanent custom configuration functions ###
+
+# determine which configuration file Apache/TestConfigData.pm to use
+# (as there could be several). The order searched is:
+# 1) $ENV{HOME}/.apache-test/
+# 2) in @INC
+my $custom_config_path;
+sub custom_config_path {
+
+    return $custom_config_path if $custom_config_path;
+
+    my @inc  = ();
+
+    # XXX $ENV{HOME} isn't propagated in mod_perl
+    push @inc, catdir $ENV{HOME}, '.apache-test' if $ENV{HOME};
+
+    push @inc, @INC;
+
+    for (@inc) {
+        my $candidate = File::Spec->rel2abs(catfile $_, CUSTOM_CONFIG_FILE);
+        next unless -e $candidate;
+        # launder for -T
+        ($candidate) = $candidate =~ /^(.*)/;
+        return $custom_config_path = $candidate;
+    }
+
+    return '';
+}
+
+# tries to nuke all occurences of custom config
+# used by things outside the A-T test suite
+sub custom_config_nuke {
+    my $cwd = fastcwd();
+
+    # 1) create a fake empty (blib/)?lib/Apache/TestConfigData.pm
+    # (don't delete it since it may mess up with MakeMaker)
+    my $path = catfile $cwd, "lib", Apache::TestConfig::CUSTOM_CONFIG_FILE;
+    # overwrite the empty stub
+    Apache::TestConfig::custom_config_write($path, '') if -e $path;
+
+    $path = catfile $cwd, "blib", "lib",
+        Apache::TestConfig::CUSTOM_CONFIG_FILE;
+    if (-e $path) {
+        my $mode = (stat _)[2];
+        my $mode_new = $mode | 0200;
+        chmod $mode_new, $path;
+        debug  "emptying $path";
+        Apache::TestConfig::custom_config_write($path, '');
+        chmod $mode, $path;
+    }
+
+    # 2) go through @INC = ~/.apache-test and nuke any occurences of
+    #    CUSTOM_CONFIG_FILE
+    my @inc  = ();
+
+    push @inc, catdir $ENV{HOME}, '.apache-test' if $ENV{HOME};
+
+    push @inc, @INC;
+
+    for (@inc) {
+        my $victim = File::Spec->rel2abs(catfile $_, CUSTOM_CONFIG_FILE);
+        next unless -e $victim;
+        debug "unlinking $victim";
+        unlink $victim;
+    }
+}
+
+sub custom_config_exists {
+    # try to load custom config if it wasn't loaded yet (there are
+    # many entry points to this API)
+    custom_config_load();
+
+    # it's enough to check whether we have a custom
+    # config for 'httpd' or 'apxs'.
+    my $httpd = $Apache::TestConfigData::vars->{httpd} || '';
+    return 1 if $httpd && -e $httpd && -x _;
+
+    my $apxs = $Apache::TestConfigData::vars->{apxs} || '';
+    return 1 if $apxs && -e $apxs && -x _;
+
+    return 0;
+}
+
+# to be used only from Apache-Test/Makefile.PL to write the custom
+# configuration module so it'll be copied to blib during 'make' and
+# updated to use custom config data during 'make test' and then
+# installed system-wide via 'make install'
+#
+# it gets written only if the custom configuration didn't exist
+# already
+sub custom_config_file_stub_write {
+
+    return if custom_config_exists();
+
+    # It doesn't matter whether it gets written under modperl-2.0/lib
+    # or Apache-Test/lib root, since Apache::TestRun uses the same
+    # logic and will update that file with real config data, which
+    # 'make install' will then pick and install system-wide. but
+    # remember that $FindBin::Bin is the location of top-level
+    # 'Makefile.PL'
+    require FindBin; # load it only for this particular use
+    my $path = catfile $FindBin::Bin, "lib",
+        Apache::TestConfig::CUSTOM_CONFIG_FILE;
+
+    # write an empty stub
+    Apache::TestConfig::custom_config_write($path, '');
+}
+
+sub custom_config_save {
+    my $self = shift;
+    my $conf_opts = shift;
+
+    if ($ENV{APACHE_TEST_NO_STICKY_PREFERENCES}) {
+        debug "APACHE_TEST_NO_STICKY_PREFERENCES=1 => " .
+            "skipping save of custom config data";
+        return;
+    }
+
+    my $vars = $self->{vars};
+    my $config_dump = '';
+
+    # minimum httpd and/or apxs needs to be set
+    return 0 unless $vars->{httpd} or $Apache::TestConfigData::vars->{httpd}
+        or          $vars->{apxs}  or $Apache::TestConfigData::vars->{apxs};
+
+    # it doesn't matter how these vars were set (httpd may or may not
+    # get set using the path to apxs, w/o an explicit -httpd value)
+    for (@data_vars_must) {
+        next unless my $var = $vars->{$_} || $conf_opts->{$_};
+        $config_dump .= qq{    '$_' => '$var',\n};
+    }
+
+    # save these vars only if they were explicitly set via command line
+    # options. For example if someone builds A-T as user 'foo', then
+    # installs it as root and we save it, all users will now try to
+    # configure under that user 'foo' which won't quite work.
+    for (@data_vars_opt) {
+        next unless my $var = $conf_opts->{$_};
+        $config_dump .= qq{    '$_' => '$var',\n};
+    }
+
+    if (IS_APACHE_TEST_BUILD) {
+        my $path = catfile $vars->{top_dir}, 'lib', CUSTOM_CONFIG_FILE;
+        # if it doesn't exist, then we already have a global config file
+        # if it does, then we have need to update it and its blib/ copy
+        if (-e $path and custom_config_path_is_writable($path)) {
+            custom_config_write($path, $config_dump);
+            # also update blib/lib, since usually that's the one that
+            # appears in @INC when t/TEST is run. and it won't be
+            # synced with blib/ unless 'make' was run
+            my $blib_path = catfile $vars->{top_dir},
+                'blib', 'lib', CUSTOM_CONFIG_FILE;
+            if (-e $blib_path and custom_config_path_is_writable($blib_path)) {
+                custom_config_write($blib_path, $config_dump);
+            }
+            return 1;
+        }
+    }
+
+    my $path;
+    if ($path = custom_config_path() ) {
+        # do nothing, the config file already exists (global)
+        debug "Found custom config '$path'";
+    }
+    elsif (File::Spec->file_name_is_absolute(__FILE__)) {
+        # next try a global location, as if it was configured before
+        # Apache::Test's 'make install' (install in the same dir as
+        # Apache/TestRun.pm)
+        # if the filename is not absolute that means that we are still
+        # in Apache-Test build (could just test for IS_APACHE_TEST_BUILD)
+        my $base = dirname dirname __FILE__;
+        $path = catdir $base, CUSTOM_CONFIG_FILE;
+    }
+
+    # check whether we can write to the directory of the chosen path
+    # (e.g. root-owned directory)
+    if ($path and custom_config_path_is_writable($path)) {
+        custom_config_write($path, $config_dump);
+        return 1;
+    }
+    # if we have no writable path yet, try to use ~
+    elsif ($ENV{HOME}) {
+        $path = catfile $ENV{HOME}, '.apache-test', CUSTOM_CONFIG_FILE;
+        if ($path and custom_config_path_is_writable($path)) {
+            custom_config_write($path, $config_dump);
+            return 1;
+        }
+    }
+
+    # XXX: should we croak since we failed to write config
+    error "Failed to find a config file to save the custom " .
+        "configuration in";
+    return 0;
+}
+
+sub custom_config_path_is_writable {
+    my $path = shift;
+
+    return 0 unless $path;
+
+    my $file_created    = '';
+    my $top_dir_created = '';
+    # first make sure that the file is writable if it exists
+    # already (it might be non-writable if installed via EU::MM or in
+    # blib/)
+    if (-e $path) {
+        my $mode = (stat _)[2];
+        $mode |= 0200;
+        chmod $mode, $path; # it's ok if we fail
+        # keep it writable if we have changed it from not being one
+        # so that custom_config_save will be able to just overwrite it
+    }
+    else {
+        my $dir = dirname $path;
+        if ($dir and !-e $dir) {
+            my @dirs = File::Path::mkpath($dir, 0, 0755);
+            # the top level dir to nuke on cleanup if it was created
+            $top_dir_created = shift @dirs if @dirs;
+        }
+        # not really create yet, but will be in the moment
+        $file_created = 1;
+    }
+
+    # try to open for append (even though it may not exist
+    my $fh = Symbol::gensym;
+    if (open $fh, ">>$path") {
+        close $fh;
+        # cleanup if we just created the file
+        unlink $path if $file_created;
+        File::Path::rmtree([$top_dir_created], 0, 0) if $top_dir_created;
+        return 1;
+    }
+
+    return 0;
+}
+
+sub custom_config_write {
+    my($path, $config_dump) = @_;
+
+    my $pkg = << "EOC";
+package Apache::TestConfigData;
+
+use strict;
+use warnings;
+
+\$Apache::TestConfigData::vars = {
+$config_dump
+};
+
+1;
+
+=head1 NAME
+
+Apache::TestConfigData - Configuration file for Apache::Test
+
+=cut
+EOC
+
+    debug "Writing custom config $path";
+    my $dir = dirname $path;
+    File::Path::mkpath($dir, 0, 0755) unless -e $dir;
+    my $fh = Symbol::gensym;
+    open $fh, ">$path" or die "Cannot open $path: $!";
+    print $fh $pkg;
+    close $fh;
+}
+
+sub custom_config_add_conf_opts {
+    my $args = shift;
+
+    return unless $Apache::TestConfigData::vars and 
+        keys %$Apache::TestConfigData::vars;
+
+    debug "overlaying custom config data";
+
+    # the logic is quite complicated with 'httpd' and 'apxs', since
+    # one is enough to run the test suite, and we need to avoid the
+    # situation where both are saved in custom config but only one
+    # (let's say httpd) is overriden by the command line /env var and
+    # a hell may break loose if we take that overriden httpd value and
+    # also use apxs from custom config which could point to a different
+    # server. So if there is an override of apxs or httpd, do not use
+    # the custom config for apxs or httpd.
+    my $vars_must_overriden = grep {
+        $ENV{ $vars_to_env{$_} } || $args->{$_}
+    } @data_vars_must;
+
+    # mod_perl 2.0 build always knows the right httpd location (and
+    # optionally apxs)
+    $vars_must_overriden++ if IS_MOD_PERL_2_BUILD();
+
+    unless ($vars_must_overriden) {
+        for (@data_vars_must) {
+            next unless $Apache::TestConfigData::vars->{$_};
+            $args->{$_} = $Apache::TestConfigData::vars->{$_};
+        }
+    }
+
+    for (@data_vars_opt) {
+        next unless $Apache::TestConfigData::vars->{$_};
+        # env vars override custom config
+        my $env_value = $ENV{ $vars_to_env{$_} };
+        next unless defined $env_value and length $env_value;
+        $args->{$_} ||= $Apache::TestConfigData::vars->{$_};
+    }
+}
+
+my $custom_config_loaded = 0;
+sub custom_config_load {
+
+    if ($ENV{APACHE_TEST_NO_STICKY_PREFERENCES}) {
+        debug "APACHE_TEST_NO_STICKY_PREFERENCES=1 => " .
+            "skipping load of custom config data";
+        return;
+    }
+
+    if ($ENV{APACHE_TEST_INTERACTIVE_CONFIG_TEST}) {
+        debug "APACHE_TEST_INTERACTIVE_CONFIG_TEST=1 => " .
+            "skipping load of custom config data";
+        return;
+    }
+
+    return if $custom_config_loaded;
+
+    if (my $custom_config_path = custom_config_path()) {
+        debug "loading custom config data from: '$custom_config_path'";
+        $custom_config_loaded++;
+        require $custom_config_path;
+    }
+    else {
+        debug "no custom config data was loaded";
+    }
+}
+
+sub custom_config_first_time {
+    my $self = shift;
+    my $conf_opts = shift;
+
+    # we can't prompt when STDIN is not attached to tty, unless we
+    # were told that's it OK via env var (in which case some program
+    # will feed the interactive prompts
+    unless (-t STDIN || $ENV{APACHE_TEST_INTERACTIVE_PROMPT_OK}) {
+        error "STDIN is not attached to tty, skip interactive config";
+        Apache::TestRun::skip_test_suite();
+    }
+
+    my $vars = $self->{vars};
+
+    print qq[
+
+We are now going to configure the Apache-Test framework.
+This configuration process needs to be done only once.
+
+];
+
+    print qq[
+
+First we need to know where the 'httpd' executable is located.
+If you have more than one Apache server is installed, make sure
+you supply the path to the one you are going to use for testing.
+You can always override this setting at run time via the '-httpd'
+option. For example:
+
+  % t/TEST -httpd /path/to/alternative/httpd
+
+or via the environment variable APACHE_TEST_HTTPD. For example:
+
+  % APACHE_TEST_HTTPD=/path/to/alternative/httpd t/TEST
+
+If for some reason you want to skip the test suite, type: skip
+];
+
+    {
+        my %choices = ();
+        my @tries = qw(httpd httpd2);
+        # Win32 uses Apache or perhaps Apache2, not apache/apache2
+        push @tries, WIN32 ? qw(Apache Apache2) : qw(apache apache2);
+        for (grep defined $_,
+             map({ catfile $vars->{$_}, $vars->{target} } qw(sbindir bindir)),
+             $self->default_httpd, which($vars->{target}),
+             $ENV{APACHE}, $ENV{APACHE2},
+             $ENV{APACHE_TEST_HTTPD}, $ENV{APACHE_TEST_HTTPD2},
+             map {which($_)} @tries) {
+            $choices{$_}++ if -e $_ && -x _;
+        }
+        my $optional = 0;
+        my $wanted = 'httpd';
+        $vars->{$wanted} = 
+            _custom_config_prompt_path($wanted, \%choices, $optional);
+    }
+
+    print qq[
+
+Next we need to know where the 'apxs' script is located. This script
+provides a lot of information about the apache installation, and makes
+it easier to find things. However it's not available on all platforms,
+therefore it's optional.
+
+If you don't have it installed it's not a problem. Just press Enter.
+
+Notice that if you have Apache 2.x installed that script could be
+called as 'apxs2'.
+
+If you have more than one Apache server is installed, make sure you
+supply the path to the apxs script you are going to use for testing.
+You can always override this setting at run time via the '-apxs'
+option. For example:
+
+  % t/TEST -apxs /path/to/alternative/apxs
+
+or via the environment variable APACHE_TEST_APXS. For example:
+
+  % APACHE_TEST_APXS=/path/to/alternative/apxs t/TEST
+
+];
+    {
+        my %choices = ();
+        for (grep defined $_,
+             map({ catfile $vars->{$_}, 'apxs' } qw(sbindir bindir)),
+             $self->default_apxs,
+             $ENV{APXS},  $ENV{APACHE_TEST_APXS},  which('apxs'),
+             $ENV{APXS2}, $ENV{APACHE_TEST_APXS2}, which('apxs2')) {
+            $choices{$_}++ if -e $_ && -x _;
+        }
+        my $optional = 1;
+        my $wanted = 'apxs';
+        $vars->{$wanted} = 
+            _custom_config_prompt_path($wanted, \%choices, $optional);
+    }
+
+    $self->custom_config_save($conf_opts);
+
+    # we probably could reconfigure on the fly ($self->configure), but
+    # the problem is various cached data which won't be refreshed. so
+    # the simplest is just to restart the run from scratch
+    Apache::TestRun::rerun($vars);
+}
+
+sub _custom_config_prompt_path {
+    my($wanted, $rh_choices, $optional) = @_;
+
+    my $ans;
+    my $default = '';
+    my $optional_str = $optional ? " (optional)" : '';
+    my $prompt =
+        "\nPlease provide a full path to$optional_str '$wanted' executable";
+
+    my @choices = ();
+    if (%$rh_choices) {
+        $prompt .= " or choose from the following options:\n\n";
+        my $c = 0;
+        for (sort keys %$rh_choices) {
+            $c++;
+            $prompt .= "    [$c] $_\n";
+            push @choices, $_;
+        }
+        $prompt .= " \n";
+        $default = 1; # a wild guess
+    }
+    else {
+        $prompt .= ":\n\n";
+    }
+
+    while (1) {
+        $ans = ExtUtils::MakeMaker::prompt($prompt, $default);
+
+        # strip leading/closing spaces
+        $ans =~ s/^\s*|\s*$//g;
+
+        # convert the item number to the path
+        if ($ans =~ /^(\d+)$/) {
+            if ($1 > 0 and $choices[$1-1]) {
+                $ans = $choices[$1-1];
+            }
+            else {
+                warn "The choice '$ans' doesn't exist\n";
+                next;
+            }
+        }
+
+        if ($optional) {
+            return '' unless $ans;
+        }
+
+        # stop the test suite without an error (so automatic tools
+        # like CPAN.pm will be able to continue)
+        if (lc($ans) eq 'skip' && !$optional) {
+            Apache::TestRun::skip_test_suite();
+            next; # in case they change their mind
+        }
+
+        unless (File::Spec->file_name_is_absolute($ans)) {
+            warn "The path '$ans' is not an absolute path. " .
+                "Please specify an absolute path\n";
+            next;
+        }
+
+        warn("'$ans' doesn't exist.\n"),     next unless -e $ans;
+        warn("'$ans' is not a file.\n"),     next unless -f _;
+        warn("'$ans' is not executable.\n"), next unless -x _;
+
+        return $ans;
+    }
+}
+
 1;
 
 =head1 NAME
@@ -1823,6 +2482,16 @@ configuration phase (C<t/TEST -config>, C<Apache::Test> will
 automatically unshift the I<project/lib> directory into C<@INC>, via
 the autogenerated I<t/conf/modperl_inc.pl> file.
 
+=head2 APACHE_TEST_INTERACTIVE_PROMPT_OK
+
+Normally interactive prompts aren't run when STDIN is not attached to
+a tty. But sometimes there is a program that can answer the prompts
+(e.g. when testing A-T itself). If this variable is true the
+interactive config won't be skipped (if needed).
+
+
+
+
 =head1 AUTHOR
 
 =head1 SEE ALSO
@@ -1859,15 +2528,17 @@ HostnameLookups Off
 </Directory>
 
 <IfModule @THREAD_MODULE@>
+    LockFile             @t_logs@/accept.lock
     StartServers         1
     MinSpareThreads      @MinClients@
     MaxSpareThreads      @MinClients@
     ThreadsPerChild      @MinClients@
-    MaxClients           @MaxClients@
+    MaxClients           @MaxClientsThreadedMPM@
     MaxRequestsPerChild  0
 </IfModule>
 
 <IfModule perchild.c>
+    LockFile             @t_logs@/accept.lock
     NumServers           1
     StartThreads         @MinClients@
     MinSpareThreads      @MinClients@
@@ -1877,6 +2548,7 @@ HostnameLookups Off
 </IfModule>
 
 <IfModule prefork.c>
+    LockFile             @t_logs@/accept.lock
     StartServers         @MinClients@
     MinSpareServers      @MinClients@
     MaxSpareServers      @MinClients@
@@ -1885,6 +2557,7 @@ HostnameLookups Off
 </IfModule>
 
 <IfDefine APACHE1>
+    LockFile             @t_logs@/accept.lock
     StartServers         @MinClients@
     MinSpareServers      @MinClients@
     MaxSpareServers      @MinClients@
@@ -1893,7 +2566,7 @@ HostnameLookups Off
 </IfDefine>
 
 <IfModule mpm_winnt.c>
-    ThreadsPerChild      20
+    ThreadsPerChild      50
     MaxRequestsPerChild  0
 </IfModule>
 
